#
# Copyright (c) 2021 Software AG, Darmstadt, Germany and/or its licensors
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Main cli interface to c8ylp"""

import click

from .cli.login import login
from .cli.plugin import cli_plugin
from .cli.server import server
from .cli.version import version
from .cli.connect import commands as connect

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


@click.group()
def cli_core():
    """Main cli entry point"""


cli_core.add_command(login)
cli_core.add_command(version)
cli_core.add_command(server)
cli_core.add_command(connect.connect)

cli = click.CommandCollection(
    "cli",
    sources=[cli_core, cli_plugin],
    context_settings=CONTEXT_SETTINGS,
    invoke_without_command=True,
    no_args_is_help=True,
    help="Cumulocity Remote Access Local Proxy",
)



#
# Copyright (c) 2021 Software AG, Darmstadt, Germany and/or its licensors
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Helper cli tests"""

from click.testing import CliRunner
from c8ylp.helper import cli, socketcontext, get_unused_port


def test_port():
    """Test get unused port number"""

    runner = CliRunner()
    result = runner.invoke(
        cli,
        ["port"],
    )
    assert result.exit_code == 0
    assert 0 <= int(result.output) <= 65535


def test_wait():
    """Test wait for port"""

    with socketcontext() as sock:
        sock.bind(("127.0.0.1", 0))
        port = sock.getsockname()[1]

        runner = CliRunner()
        result = runner.invoke(
            cli,
            ["wait", str(port), "5"],
        )
        assert result.exit_code == 0


def test_wait_timout_error():
    """Test timeout error whilst waiting for a port"""

    runner = CliRunner()
    port = get_unused_port()
    result = runner.invoke(
        cli,
        ["wait", str(port), "0.5"],
    )
    assert result.exit_code == 1
    assert "Port is not open after 0.5s" in result.output


def test_wait_silent_error():
    """Test wait timeout error but skip stdout message and only set exit code"""

    runner = CliRunner()
    port = get_unused_port()
    result = runner.invoke(
        cli,
        ["wait", str(port), "0.5", "--silent"],
    )
    assert result.exit_code == 1
    assert "Port is not open after" not in result.output



#
# Copyright (c) 2021 Software AG, Darmstadt, Germany and/or its licensors
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""ssh. Used to start an interactive ssh session or execute a single ssh
command then exit.
"""

import logging
import os
import shutil
import subprocess
from typing import List

import click
from c8ylp import options
from c8ylp.cli.core import ExitCodes, ProxyContext


@click.command()
@options.common_options
@click.option(
    "--ssh-user",
    required=True,
    envvar="C8YLP_SSH_USER",
    show_envvar=True,
    prompt=True,
    help="SSH username which is configured on the device",
)
@click.argument(
    "additional_args", metavar="[REMOTE_COMMANDS]...", nargs=-1, type=click.UNPROCESSED
)
@click.pass_context
def cli(ctx: click.Context, ssh_user: str, additional_args: List[str], **kwargs):
    """Start once-off proxy and connect via ssh

    An interactive ssh is opened if a remote command is not provided.

    Once the ssh session is closed, then the local proxy will be shutdown.

    Use "--" before the remote commands to prevent the arguments
    from being interpreted by c8ylp (i.e. to avoid clashes with c8ylp).

    \b
        DEVICE is the device's external identity

    Example 1: Start an interactive SSH connection

    \b
        c8ylp connect ssh device01 --env-file .env --ssh-user admin

    Example 2: Execute a command via SSH

    \b
        c8ylp connect ssh device01 --env-file .env --ssh-user admin -- systemctl status ssh

    Example 3: Execute a complex command via SSH (use quotes to ensure command is sent to the device)

    \b
        c8ylp connect ssh device01 --env-file .env --ssh-user admin -- "systemctl status ssh; dpkg --list | grep ssh"

    """
    proxy = ProxyContext(ctx, kwargs).start_background()

    if not shutil.which("ssh"):
        proxy.show_error(
            "ssh client not found. Please make sure the 'ssh' client is included in your PATH variable"
        )
        ctx.exit(ExitCodes.SSH_NOT_FOUND)

    device = proxy.device
    host = proxy.host

    ssh_args = [
        "ssh",
        "-o",
        "ServerAliveInterval=120",
        "-o",
        "StrictHostKeyChecking=no",
        "-o",
        "UserKnownHostsFile=/dev/null",
        "-p",
        str(proxy.used_port),
        f"{ssh_user}@localhost",
    ]

    if additional_args:
        ssh_args.extend(additional_args)
        proxy.show_message(f"Executing command via ssh on {device} ({host})")
    else:
        proxy.show_message(f"Starting interactive ssh session with {device} ({host})")

    logging.info("Starting ssh session using: %s", " ".join(ssh_args))
    exit_code = subprocess.call(ssh_args, env=os.environ)
    if exit_code != 0:
        proxy.show_error(f"SSH exited with a non-zero exit code. code={exit_code}")
    ctx.exit(exit_code)


if __name__ == "__main__":
    # pylint: disable=no-value-for-parameter
    cli()



#
# Copyright (c) 2021 Software AG, Darmstadt, Germany and/or its licensors
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Test plugin command"""

import logging
import subprocess
import sys
import time
import threading
from typing import List

import pytest

from .fixtures import Device


def proxy_cli(*args) -> subprocess.CompletedProcess:
    """Execute the proxy cli command with given arguments"""
    return subprocess.run(
        [
            sys.executable,
            "-m",
            "c8ylp",
            *args,
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )


@pytest.mark.parametrize(
    "case",
    (
        dict(clients=1, delay=10, command="sleep 10"),
        dict(clients=2, delay=10, command="sleep 10"),
        dict(clients=5, delay=7, command="sleep 10"),
    ),
    ids=lambda x: str(x),
)
def test_concurrent_commands(case, c8ydevice: Device):
    """Test running concurrent commands"""
    clients = case.get("clients")
    delay = case.get("delay", 0.1)
    command = case.get("command", "sleep 10")

    threads: List[threading.Thread] = []
    results: List[subprocess.CompletedProcess] = [None] * clients
    durations: List[float] = [None] * clients

    for i in range(0, clients):

        cmd_args = (
            "plugin",
            "command",
            c8ydevice.device,
            "--",
            command,
        )

        def run_proxy(index, *args):
            start = time.monotonic()
            results[index] = proxy_cli(*args)
            durations[index] = time.monotonic() - start
            logging.info(
                f"client={index+1}, exit_code={results[index].returncode} duration={int(durations[index])}"
            )

        thread = threading.Thread(target=run_proxy, args=[i, *cmd_args], daemon=True)
        thread.start()
        threads.append(thread)

        if i < clients - 1:
            time.sleep(delay)

    # wait for threads to finish
    for thread in threads:
        thread.join()

    assert [proc.returncode for proc in results] == [0] * clients



#
# Copyright (c) 2021 Software AG, Darmstadt, Germany and/or its licensors
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Integration test fixtures"""

import os
from .fixtures import Device, FileFactory
from c8ylp.env import loadenv

import pytest


@pytest.fixture
def c8ydevice():
    """Cumulocity integration test device fixture"""

    test_env = os.getenv("TEST_ENV_FILE", ".env")
    os.environ.setdefault("C8YLP_ENV_FILE", test_env)

    if os.path.exists(test_env):
        loadenv(test_env)

    yield Device(
        device=os.getenv("TEST_DEVICE", ""), ssh_user=os.getenv("TEST_SSH_USER", "")
    )


@pytest.fixture
def file_factory(tmpdir):
    """File factory fixture"""
    yield FileFactory(dir=tmpdir)



#
# Copyright (c) 2021 Software AG, Darmstadt, Germany and/or its licensors
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Integration fixtures"""

import os
from dataclasses import dataclass
from pathlib import Path


@dataclass
class Device:
    device: str = ""
    ssh_user: str = ""


class FileFactory:
    """File factory to generate files with random content"""

    def __init__(self, dir: str) -> None:
        self._dir = Path(dir)

    @property
    def dir(self) -> Path:
        """Directory where the temp files are created"""
        return self._dir

    def create_file(self, name: str, size_mb: int = 10) -> Path:
        path = self._dir / name
        with open(path, "wb") as file:
            for i in range(0, size_mb * 1000):
                file.write(os.urandom(1024))

        return path



#
# Copyright (c) 2021 Software AG, Darmstadt, Germany and/or its licensors
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Plugin run"""

import platform
import os
import stat
from pathlib import Path

import pytest

import responses
from click.testing import CliRunner
from c8ylp.main import cli
from c8ylp.cli.core import ExitCodes
from tests.env import Environment
from tests.fixtures import FixtureCumulocityAPI


def create_plugin(path: Path, code: str) -> str:
    """Format code to be unix compatible (i.e. strip \\r characters)
    Note: Path.write_text does not work as it does not allow to set the
    newline character (but this seems to be fixed in Python 3.10)

    Args:
        path (Path): Path to the script to be written to
        code (str): Code to be formatted

    Returns:
        str: Unix friendly code
    """
    with path.open("w", newline="\n", encoding="utf8") as file:
        file.write(code.replace("\r", "", -1))


@pytest.fixture(name="plugin")
def plugin_fixture(tmpdir):
    """Proxy plugin fixture which creates a temporary plugin folder"""
    plugin_folder = Path(tmpdir) / ".c8ylp" / "plugins"
    plugin_folder.mkdir(parents=True, exist_ok=True)
    yield plugin_folder


def test_plugin_list(
    plugin: Path,
):
    """Test getting a list of the plugins"""

    # Simulate some plugins
    bash_plugin_1 = Path(plugin / "launch.sh")
    bash_plugin_1.touch()

    bash_plugin_2 = Path(plugin / "scp.sh")
    bash_plugin_2.touch()

    py_plugin_1 = Path(plugin / "mycommand.sh")
    py_plugin_1.touch()

    runner = CliRunner()
    result = runner.invoke(
        cli,
        ["plugin"],
        env={
            "C8YLP_PLUGINS": str(plugin),
        },
    )

    assert result.exit_code == ExitCodes.OK
    subcommands = []
    found = False

    for line in result.output.splitlines():
        if found:
            name, _, _ = line.strip().partition(" ")
            subcommands.append(name)
        if "Commands:" in line:
            found = True

    assert subcommands == ["command", "launch", "mycommand", "scp"]


def test_calling_unknown_plugin(
    plugin: Path,
):
    """Test calling an unknown plugin"""

    runner = CliRunner()
    result = runner.invoke(
        cli,
        ["plugin", "doesnotexist"],
        env={
            "C8YLP_PLUGINS": str(plugin),
        },
    )

    assert result.exit_code == ExitCodes.PLUGIN_NOT_FOUND


def test_bash_plugin(
    plugin: Path,
    c8yserver: FixtureCumulocityAPI,
    env: Environment,
    capfd: pytest.CaptureFixture,
):
    """Test bash plugin"""

    @responses.activate
    def run():
        serial = "ext-device-01"
        port = "2223"
        c8yserver.simulate_pre_authenticated(serial)

        bash_plugin_1 = Path(plugin / "launch.sh")

        create_plugin(
            bash_plugin_1,
            """
        #!/bin/bash
        echo "Running my custom launcher: PORT=$PORT, DEVICE=$DEVICE"
        """.lstrip(),
        )

        # make plugin executable
        exiting_stat = os.stat(bash_plugin_1)
        os.chmod(bash_plugin_1, exiting_stat.st_mode | stat.S_IEXEC)

        runner = CliRunner()
        result = runner.invoke(
            cli,
            [
                "plugin",
                "launch",
                "--port",
                port,
                serial,
            ],
            env={
                **env.create_authenticated(),
                "C8YLP_PLUGINS": str(plugin),
            },
        )

        assert result.exit_code == ExitCodes.OK

        out, _ = capfd.readouterr()
        assert out == f"Running my custom launcher: PORT={port}, DEVICE={serial}\n"

    run()


@pytest.mark.skipif(
    platform.system() == "Windows", reason="Windows does not throw an error"
)
def test_invalid_bash_plugin(
    plugin: Path,
    c8yserver: FixtureCumulocityAPI,
    env: Environment,
):
    """Test invalid bash plugin"""

    @responses.activate
    def run():
        serial = "ext-device-01"
        port = "2224"
        c8yserver.simulate_pre_authenticated(serial)

        bash_plugin_1 = Path(plugin / "launch.sh")

        create_plugin(
            bash_plugin_1,
            """

        # Plugin which is missing the shebang!
        echo "Running my customer launcher: PORT=$PORT"
        """,
        )

        # make plugin executable
        existing_stat = os.stat(bash_plugin_1)
        os.chmod(bash_plugin_1, existing_stat.st_mode | stat.S_IEXEC)

        runner = CliRunner()
        result = runner.invoke(
            cli,
            [
                "plugin",
                "launch",
                "--port",
                port,
                serial,
            ],
            env={
                **env.create_authenticated(),
                "C8YLP_PLUGINS": str(plugin),
            },
        )
        assert result.exit_code == ExitCodes.PLUGIN_EXECUTION_ERROR

    run()


@pytest.mark.parametrize(
    "code,exit_code",
    [
        (
            """
import click
    @click.command()
    def cli():
        print("Running plugin")
        """,
            ExitCodes.PLUGIN_INVALID_FORMAT,
        ),
        (
            """
import click
@click.command()
def invalid_name():
    print("Running plugin")
        """,
            ExitCodes.PLUGIN_INVALID_FORMAT,
        ),
        (
            """
import click
from c8ylp import options

@click.command()
@options.common_options
def cli(**kwargs):
    print("Running plugin")
        """,
            ExitCodes.OK,
        ),
    ],
    ids=["invalid indentation", "wrong func name", "OK"],
)
def test_invalid_python_plugin(
    code: str,
    exit_code: int,
    plugin: Path,
    c8yserver: FixtureCumulocityAPI,
    env: Environment,
):
    """Test invalid python plugin"""

    @responses.activate
    def run():
        serial = "ext-device-01"
        c8yserver.simulate_pre_authenticated(serial)

        plugin_1 = Path(plugin / "launch.py")
        create_plugin(plugin_1, code)

        runner = CliRunner()
        result = runner.invoke(
            cli,
            [
                "plugin",
                "launch",
                serial,
            ],
            env={
                **env.create_authenticated(),
                "C8YLP_PLUGINS": str(plugin),
            },
        )
        assert result.exit_code == exit_code

    run()



#
# Copyright (c) 2021 Software AG, Darmstadt, Germany and/or its licensors
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""SSH command tests"""
import os
from unittest.mock import Mock, patch

import pytest
from c8ylp.rest_client.c8yclient import CumulocityClient
from tests.env import Environment
from tests.fixtures import FixtureCumulocityAPI, LocalProxyLog


@pytest.fixture(name="c8yclient")
def fixture_c8yclient():
    """Cumulocity client fixture"""
    dummy_client = CumulocityClient("https://example.c8y.io")

    def set_tenant():
        dummy_client.tenant = "t12345"
        return dummy_client.tenant

    dummy_client.validate_tenant_id = Mock(side_effect=set_tenant)

    with patch("c8ylp.main.CumulocityClient", return_value=dummy_client) as mock_client:
        yield mock_client


@pytest.fixture(name="env")
def fixture_env():
    """Environment fixture"""
    yield Environment()


@pytest.fixture(name="c8yserver")
def fixture_c8yserver():
    """Cumulocity server fixture to simulate server responses"""
    api = FixtureCumulocityAPI()
    yield api


@pytest.fixture(name="localproxy_log")
def fixture_localproxy_log():
    """Local proxy log fixture"""
    log = LocalProxyLog()
    yield log


@pytest.fixture(autouse=True)
def run_before_and_after_tests(localproxy_log: LocalProxyLog, tmpdir):
    """Fixture to execute asserts before and after a test is run"""
    # Setup: Clear any existing logs
    os.environ["C8YLP_LOG_DIR"] = str(tmpdir)

    yield

    # Show log log output
    localproxy_log.print_output()



#
# Copyright (c) 2021 Software AG, Darmstadt, Germany and/or its licensors
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Timer"""

import logging
import time
from datetime import timedelta
from typing import Callable


class CommandTimer:
    """Command Timer which shows how long a command takes to run
    and prints out a message to the user

    Example

    >>>
    with CommandTimer():
        print("Doing someting")
        time.sleep(100)
    >>>

    """

    def __init__(self, message: str, on_exit: Callable[[], None] = None) -> None:
        self.message = message
        self.start_time = 0
        self.last_duration = 0
        self._on_exit = on_exit

    def start(self):
        """Start the timer"""
        self.start_time = time.monotonic()
        return self

    def stop(self) -> float:
        """Stop the timer and return the duration in seconds

        Returns:
            float: Duration in seconds
        """
        if not self.start_time:
            return 0
        self.last_duration = time.monotonic() - self.start_time
        return self.last_duration

    def stop_with_message(self):
        """Stop timer and print out a message about the duration"""
        duration = timedelta(seconds=(round(self.stop(), 0)))
        msg = f"{self.message}: {duration}"
        logging.info(msg)
        if callable(self._on_exit):
            self._on_exit(msg)

    def __enter__(self) -> None:
        self.start()

    def __exit__(self, _type, _value, _traceback) -> None:
        self.stop_with_message()



#
# Copyright (c) 2021 Software AG, Darmstadt, Germany and/or its licensors
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Test plugin shell handling"""

import pytest

from c8ylp.cli.plugin import format_wsl_path


@pytest.mark.parametrize(
    "path,expected",
    [
        (
            "c:\\myscript\\ to my\\plugin.sh",
            "/mnt/c/myscript/ to my/plugin.sh",
        ),
        (
            "D:\\launcher.sh",
            "/mnt/d/launcher.sh",
        ),
    ],
)
def test_format_wsl_path(path, expected):
    """Test formating of a windows path to the WSL equivalent"""
    assert format_wsl_path(path) == expected



#
# Copyright (c) 2021 Software AG, Darmstadt, Germany and/or its licensors
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""version command"""

import click

from .. import __version__
from .core import ExitCodes


@click.command()
@click.pass_context
def version(ctx: click.Context):
    """Show the c8ylp version number"""
    click.echo(f"Version {__version__}")
    ctx.exit(ExitCodes.OK)
